@@ -1,5 +1,5 @@
 package ExtUtils::Typemaps::Magic;
-$ExtUtils::Typemaps::Magic::VERSION = '0.002';
+$ExtUtils::Typemaps::Magic::VERSION = '0.003';
 use strict;
 use warnings;
 
@@ -10,9 +10,9 @@ sub new {
 
 	my $self = $class->SUPER::new(@_);
 
-	$self->add_inputmap(xstype => 'T_MAGIC', code => <<'END');
+	my $input = <<'END';
 	{
-	MAGIC* magic = SvROK($arg) && SvMAGICAL(SvRV($arg)) ? mg_findext(SvRV($arg), PERL_MAGIC_ext, NULL) : NULL;
+	MAGIC* magic = SvROK($arg) && SvMAGICAL(SvRV($arg)) ? mg_findext(SvRV($arg), PERL_MAGIC_ext, %s) : NULL;
 	if (magic)
 		$var = (${type})magic->mg_ptr;
 	else
@@ -20,17 +20,12 @@ sub new {
 	}
 END
 
-	$self->add_inputmap(xstype => 'T_MAGICEXT', code => <<'END');
-	{
-	MAGIC* magic = SvROK($arg) && SvMAGICAL(SvRV($arg)) ? mg_findext(SvRV($arg), PERL_MAGIC_ext, &${type}_magic) : NULL;
-	if (magic)
-		$var = (${type})magic->mg_ptr;
-	else
-		Perl_croak(aTHX_ \"${ntype} object is lacking magic\");
-	}
-END
+	$self->add_inputmap(xstype => 'T_MAGIC', code => sprintf $input, 'NULL');
+	$self->add_inputmap(xstype => 'T_MAGICBUF', code => sprintf $input, 'NULL');
+	$self->add_inputmap(xstype => 'T_MAGICEXT', code => sprintf $input, '&${type}_magic');
 
 	$self->add_outputmap(xstype => 'T_MAGIC', code => '	sv_magic(newSVrv($arg, \"${ntype}\"), NULL, PERL_MAGIC_ext, (const char*)$var, 0);');
+	$self->add_outputmap(xstype => 'T_MAGICBUF', code => '	sv_magic(newSVrv($arg, \"${ntype}\"), NULL, PERL_MAGIC_ext, (const char*)$var, sizeof(*$var));');
 
 	$self->add_outputmap(xstype => 'T_MAGICEXT', code => <<'END');
 	{
@@ -62,7 +57,7 @@ ExtUtils::Typemaps::Magic - Typemap for storing objects in magic
 
 =head1 VERSION
 
-version 0.002
+version 0.003
 
 =head1 SYNOPSIS
 
@@ -79,31 +74,33 @@ version 0.002
 
 =head1 DESCRIPTION
 
-C<ExtUtils::Typemaps::Magic> is an C<ExtUtils::Typemaps> subclass that provides two magic based mappings for objects.
+C<ExtUtils::Typemaps::Magic> is an C<ExtUtils::Typemaps> subclass that provides three magic based mappings for objects:
 
 =head2 T_MAGIC
 
-This is essentially a drop-in replacement for C<T_PTROBJ>, except that it hides the value of the pointer from pure-perl code by storing it in attached magic. In particular that means the pointer won't be serialized/deserialized (this is usually a thing because after deserialization the pointer is probably not valid).
+This is essentially a drop-in replacement for C<T_PTROBJ>, except that it hides the value of the pointer from pure-perl code by storing it in attached magic. In particular that means the pointer won't be serialized/deserialized (this is usually a thing because after deserialization the pointer is probably not valid). Note that like C<T_PTROBJ>, you probably need a C<DESTROY> method to destroy and free the buffer. Like C<T_PTROBJ> and friends, this is not thread cloning safe without further measures.
+
+=head2 T_MAGICBUF
+
+This is equivalent of using a string reference to store your object in, except it is hidden away using magic. This is suitable for objects that can be safely shallow copied on thread cloning (i.e. they don't contain external references such as pointers or file descriptors). Unlike C<T_MAGIC> or C<T_PTROBJ> this does not need a C<DESTROY> method to free the buffer.
 
 =head2 T_MAGICEXT
 
 This stores the object just like C<T_MAGIC> does, but additionally attaches a magic vtable (type C<MGVTBL>) with the name C<${type}_magic> (e.g. C<Foo__Bar_magic> for a value of type C<Foo::Bar>) to the value. This is mainly useful for adding C<free> (destruction) and C<dup> (thread cloning) callbacks. The details of how these work is explained in L<perlguts|perlguts>, but it might look something like this:
 
  static int object_dup(pTHX_ MAGIC* magic, CLONE_PARAMS* params) {
-     PERL_UNUSED_VAR(params);
      object_refcount_increment((struct Object*)magic->mg_ptr);
      return 0;
  }
 
  static int object_free(pTHX_ SV* sv, MAGIC* magic) {
-     PERL_UNUSED_VAR(sv);
      object_refcount_decrement((struct Object*)magic->mg_ptr);
      return 0;
  }
 
  static const MGVTBL My__Object_magic = { NULL, NULL, NULL, NULL, object_free, NULL, object_dup, NULL };
 
-This is useful to create objects that handle thread cloning correctly and effectively.
+This is useful to create objects that handle thread cloning correctly and effectively. The object should be allocated with the C<PerlSharedMem_malloc> family of allocators.
 
 =head1 DEPENDENCIES
 
